% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 7: Basic Corpus Statistics

disp('Example 7.1 ------------------------------------------------------');
disp('THIS EXAMPLE IS NOT A MATLAB SCRIPT')                          %(7.1)
pause;

disp('Example 7.2 ------------------------------------------------------');
clear;                                                              %(7.2a)
collection = string(evalc('type bible_en.xml'));
whos collection                                                     %(7.2b)
pause;

disp('Example 7.3 ------------------------------------------------------');
collection = regexprep(collection,'<.*?>','');                       %(7.3)
pause;

disp('Example 7.4 ------------------------------------------------------');
collection = lower(collection);                                      %(7.4)
pause;

disp('Example 7.5 ------------------------------------------------------');
collection = regexprep(collection,'\W',' ');                         %(7.5)
pause;

disp('Example 7.6 ------------------------------------------------------');
collection = strtrim(collection);                                    %(7.6)
pause;

disp('Example 7.7 ------------------------------------------------------');
words = split(collection);                                           %(7.7)
pause;

disp('Example 7.8 ------------------------------------------------------');
running_words = length(words)                                        %(7.8)
pause;

disp('Example 7.9 ------------------------------------------------------');
vocabulary = unique(words);                                         %(7.9a)
vocabulary_words = length(vocabulary)                               %(7.9b)
pause;

disp('Example 7.10 -----------------------------------------------------');
[vocabulary,void,index] = unique(words);                           %(7.10a)
whos index
frequencies = hist(index,vocabulary_words);                        %(7.10b)
whos frequencies
pause;

disp('Example 7.11 -----------------------------------------------------');
[ranked_frequencies,ranking_index] = sort(frequencies,'descend');  %(7.11a)
ranked_vocabulary = vocabulary(ranking_index);  	               %(7.11b)
pause;

disp('Example 7.12 -----------------------------------------------------');
set(figure,'Color',[1 1 1],'Name','Zipf''s Law');                  %(7.12a)
loglog(ranked_frequencies,'.');                                    %(7.12b)
xlabel('Rank'); ylabel('Word frequency');                          %(7.12c)
pause;

disp('Example 7.13 -----------------------------------------------------');
ranked_vocabulary(1:3)'                                             %(7.13)
pause;

disp('Example 7.14 -----------------------------------------------------');
sum(ranked_frequencies(1:3))                                        %(7.14)
pause;

%%%%%%%%%%%%%%%%%%% Table 6.1
%n=7; 
%x=sum(ranked_frequencies(1:n));
%fprintf('%3.2f\n%d\n%4.2f\n',n/vocabulary_words*100,x,x/running_words*100)
%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 7.15 -----------------------------------------------------');
disp('THIS EXAMPLE IS NOT A MATLAB SCRIPT')                         %(7.15)
pause;

disp('Example 7.16 -----------------------------------------------------');
estimated_probabilities = ranked_frequencies/running_words;  	    %(7.16)
pause;

disp('Example 7.17 -----------------------------------------------------');
sum(estimated_probabilities)  	                                    %(7.17)
pause;

disp('Example 7.18 -----------------------------------------------------');
estimated_probabilities(strcmp(ranked_vocabulary,'lord'))   	   %(7.18a)
estimated_probabilities(strcmp(ranked_vocabulary,'water'))   	   %(7.18b)
estimated_probabilities(strcmp(ranked_vocabulary,'taxes'))   	   %(7.18c)
pause;

disp('Example 7.19 -----------------------------------------------------');
estimated_probabilities(strcmp(ranked_vocabulary,'complexity'))  	%(7.19)
pause;

disp('Example 7.20 -----------------------------------------------------');
ranked_vocabulary(1:30)'  	                                        %(7.20)
pause;

disp('Example 7.21 -----------------------------------------------------');
ranked_vocabulary(201:230)'  	                                    %(7.21)
pause;

disp('Example 7.22 -----------------------------------------------------');
mean(strlength(ranked_vocabulary(1:30)))                           %(7.22a)
mean(strlength(ranked_vocabulary(201:230)))                        %(7.22b)
mean(strlength(ranked_vocabulary(1001:1030)))                      %(7.22c)
mean(strlength(ranked_vocabulary(11001:11030)))                    %(7.22d)
pause;

%%%%%%%%%%%%%%%%%%%%%%% old way
%sum(sum(not(char(ranked_vocabulary(1:30))==' ')))/30  	           %(6.22a)
%sum(sum(not(char(ranked_vocabulary(201:230))==' ')))/30  	       %(6.22b)
%sum(sum(not(char(ranked_vocabulary(1001:1030))==' ')))/30  	       %(6.22c)
%sum(sum(not(char(ranked_vocabulary(11001:11030))==' ')))/30  	   %(6.22d)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 7.23 -----------------------------------------------------');
hf = figure(2);  	                                               %(7.23a)
set(hf,'Color',[1 1 1],'Name','Another Zipf''s Law');
semilogx(strlength(ranked_vocabulary),'.');  	                   %(7.23b)
xlabel('Rank'); ylabel('Word length');  	                       %(7.23c)
pause;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%semilogx(sum(not(char(ranked_vocabulary)==' '),2),'.');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5555

disp('Example 7.24 -----------------------------------------------------');
running_words/ranked_frequencies(strcmp(ranked_vocabulary,'the'))   %(7.24)
pause;

disp('Example 7.25 -----------------------------------------------------');
% gets the indexes for the occurrences of the word 'the'  	       %(7.25a)
locations = find(strcmp(words,'the'));
% computes the lengths of the repetition intervals  	           %(7.25b)
intervals = diff(locations);
% computes a 100-bin histogram  	                               %(7.25c)
histogram = hist(intervals,100);
hf = figure(3); % creates a new figure  	                       %(7.25d)
set(hf,'Color',[1 1 1],'Name','Intermittency property');
% plots interval lengths between consecutive repetitions of 'the'  %(7.25e)
subplot(2,1,1); nvals = length(intervals);
plot(1:nvals,intervals,'-r',[1,nvals],ones(1,2)*mean(intervals),'--k');
limits = axis; axis([1,nvals,limits(3:4)]); ylabel('Interval length');
xlabel('Consecutive occurrences of word ''the'''); 
% plots the histogram of interval lengths  	                       %(7.25f)
subplot(2,2,3); bar(histogram);
xlabel('Interval length'); ylabel('Frequency');
% plots the distribution (histogram) in logarithmic space  	       %(7.25g)
subplot(2,2,4); loglog(1:length(histogram),histogram,'.k');
xlabel('Interval length'); ylabel('Frequency');
pause;

disp('Example 7.26 -----------------------------------------------------');
median(intervals)                                                   %(7.26)
pause;

disp('Example 7.27 -----------------------------------------------------');
% writes words 'water' and 'child' into a cell array  	           %(7.27a)
twowords = {'water','child'};
% retrieves their ranking positions  	                           %(7.27b)
find(strcmp(ranked_vocabulary,twowords{1}))
find(strcmp(ranked_vocabulary,twowords{2}))
hf = figure(4); % creates a new figure  	                       %(7.27c)
set(hf,'Color',[1 1 1],'Name','Two more interval length plots');
% generates the plots  	                                           %(7.27d)
for k=1:2
    intervals = diff(find(strcmp(words,twowords{k})));
    subplot(2,1,k); nvals = length(intervals);
    plot(1:nvals,intervals,'-r',[1,nvals],ones(1,2)*mean(intervals),'--k');       
    limits = axis; axis([1,nvals,limits(3:4)]);
    xlabel(sprintf('Consecutive occurrences of word ''%s''',twowords{k}));
    ylabel('Interval length');
end;
pause;

disp('Example 7.28 -----------------------------------------------------');
temp = string(evalc('type bible_en.xml')); % reads the file 	   %(7.28a)
temp = lower(temp); % lowercases the text	                       %(7.28b)
% inserts markers at the end of each verse before removing the tags%(7.28c)
temp = regexprep(temp,'</verse>',' S '); 
temp = regexprep(temp,'<.*?>',''); % removes the tags	           %(7.28d)
temp = regexprep(temp,'\W',' '); % eliminates non-word characters  %(7.28e)
temp = strtrim(temp); % removes leading and trailing whitespaces   %(7.28f)
% segments the data collection into tokens	                       %(7.28g)
wordsofverses = split(temp);
% locates the verse boundaries	                                   %(7.28h)
limits = [0,find(strcmp(wordsofverses','S'))];
% creates the structure array	                                   %(7.28i)
for k=1:length(limits)-1   
    verses(k).vocab = unique(wordsofverses(limits(k)+1:limits(k+1)-1));
end; 
pause;

disp('Example 7.29 -----------------------------------------------------');
% creates a sparse matrix of appropriate dimensions  	           %(7.29a)
tdmtx = sparse(length(ranked_vocabulary),length(verses));
% updates the corresponding term-document matrix elements  	       %(7.29b)
for k=1:length(verses) % this will take a while...
    [void,termindex] = intersect(ranked_vocabulary,verses(k).vocab);
    tdmtx(termindex,k) = 1;
end;
pause;

disp('Example 7.30 -----------------------------------------------------');
comtx = tdmtx*tdmtx';   	                                        %(7.30)
pause;

disp('Example 7.31 -----------------------------------------------------');
sum(sum(comtx>0))/numel(comtx)*100                                  %(7.31)
pause;

disp('Example 7.32 -----------------------------------------------------');
disp('THIS EXAMPLE IS NOT A MATLAB SCRIPT')                         %(7.32)
pause;

disp('Example 7.33 -----------------------------------------------------');
p1word_verse = diag(comtx)/length(verses);                          %(7.33)
pause;

disp('Example 7.34 -----------------------------------------------------');
[wa,wb,rawcount] = find(triu((comtx>30).*comtx,1));   	           %(7.34a)
p2word_verse = rawcount/length(verses);   	                       %(7.34b)
pause;

disp('Example 7.35 -----------------------------------------------------');
minfo_verse = p2word_verse./(p1word_verse(wa).*p1word_verse(wb));   %(7.35)
pause;

disp('Example 7.36 -----------------------------------------------------');
% sorts mutual information scores in descending order  	           %(7.36a)
[sortminfo,sortindex] = sort(minfo_verse,'descend');
% copies all wa into a string array according to sortindex order   %(7.36b)
tempa = ranked_vocabulary(wa(sortindex));
% copies all wb into a string array according to sortindex order   %(7.36c)
tempb = ranked_vocabulary(wb(sortindex));
% creates a table containing words and mutual information scores   %(7.36d)
minfo_rank = table;
minfo_rank.WORDa = tempa;
minfo_rank.WORDb = tempb;
minfo_rank.PMI_Score = sortminfo;
% lists the first ten values for tempa, tempb and sortminfo   	   %(7.36e)
disp('-> Pairs of words with largest pointwise mutual information');
disp(head(minfo_rank,10));
pause;

disp('Example 7.37 -----------------------------------------------------');
% lists the last ten values for tempa, tempb and sortminfo	        %(7.37)
disp('-> Pairs of words with lowest pointwise mutual information');
disp(tail(minfo_rank,10));
pause;

disp('Example 7.38 -----------------------------------------------------');
% sorts mutual information scores as they depart from the unit	   %(7.38a)
[void,absindex] = sort(abs(sortminfo-1));
% lists values for tempa, tempb and sortminfo closest to the unit  %(7.38b)
disp('-> Pairs of words with pointwise mutual information closest to 1.0');
disp(head(minfo_rank(absindex,:),10));
pause;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Table 7.2
%wlist1 = ["purple","sell","tree","wife","more","your","water",...
%          "fruit","thy","house","say","soon","for","water",...
%          "tell","thou","house","thy","will","had"];
%wlist2 = ["blue","buy","fruit","husband","less","you","river",...
%          "seed","thou","wife","word","there","well","house",...
%          "word","you","tree","your","saw","shall"];
%widx1 = []; widx2 =[];
%for k=1:length(wlist1)
%    widx1(k) = find(strcmp(wlist1(k),ranked_vocabulary));
%    widx2(k) = find(strcmp(wlist2(k),ranked_vocabulary));
%end
%pa = full(p1word_verse(widx1));
%pb = full(p1word_verse(widx2));
%table2 = table;
%table2.word_a = wlist1';
%table2.word_b = wlist2';
%table2.pwa = pa;
%table2.pwb = pb;
%pjoin = [];
%for k=1:length(wlist1)
%    pjoin(k) = full(comtx(widx1(k),widx2(k))/length(verses));
%end
%table2.pwawb = pjoin';
%table2.pwacwb = pjoin'./pb;
%table2.pwbcwa = pjoin'./pa;
%table2.pmi = pjoin'./(pa.*pb);
%pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 7.39 -----------------------------------------------------');
% creates a sparse matrix for accumulating co-occurrences	       %(7.39a)
co2mtx = sparse(length(ranked_vocabulary),length(ranked_vocabulary));
% creates a vector for accumulating individual word counts	       %(7.39b)
w1raw = zeros(1,length(ranked_vocabulary));
for k=2:length(wordsofverses) % moves along the dataset  	       %(7.39c)
    % gets the next two consecutive words 
    w1 = wordsofverses(k-1); w2 = wordsofverses(k);  
    % avoids any pair of words containing the end-of-verse marker 'S'
    if w1~='S' && w2~='S'
        % gets the rank indexes of w1 and w2 
        idx1 = find(ranked_vocabulary==w1);
        idx2 = find(ranked_vocabulary==w2);
        % increments the counts for word sequence w1,w2 
        co2mtx(idx1,idx2) = co2mtx(idx1,idx2)+1;
        % increments the counts for single words
        w1raw(idx2) = w1raw(idx2)+1;      
        % considers the 1st word 
        %if k==2, w1raw(idx1) = w1raw(idx1)+1; end;
    end; 
end; % this will take a while...
pause;

disp('Example 7.40 -----------------------------------------------------');
p1word_next = w1raw'/(length(words)-length(verses));   	            %(7.40)
pause;

disp('Example 7.41 -----------------------------------------------------');
[w1,w2,rawcount] = find((co2mtx>30).*co2mtx);   	               %(7.41a)
p2word_next = rawcount/(length(words)-length(verses));   	       %(7.41b)
pause;

disp('Example 7.42 -----------------------------------------------------');
minfo_next = p2word_next./(p1word_next(w1).*p1word_next(w2));   	%(7.42)
pause;

%%%%%%%%% Table 7.3
% [sortminfo,sortindex] = sort(minfo_next,'descend');
% temp1 = ranked_vocabulary(w1(sortindex));
% temp2 = ranked_vocabulary(w2(sortindex));
% minfo_rank = table;
% minfo_rank.WORD1 = temp1;
% minfo_rank.WORD2 = temp2;
% minfo_rank.PMI_Score = sortminfo;
% disp('-> Largest PMI');
% disp(head(minfo_rank,10));
% disp('-> Lowest PMI');
% disp(tail(minfo_rank,10));
% [void,absindex] = sort(abs(sortminfo-1));
% disp('-> Closest to the unit');
% disp(head(minfo_rank(absindex,:),10));
%pause;
%%%%%%%%%%%%%%%%%%%

disp('Example 7.43 -----------------------------------------------------');
% defines four reference words and their corresponding probes      %(7.43a)
refer = ["father","shall","go","go"]; 
probe = ["house","be","to","for"];
hf = figure(5); % creates a figure and sets its basic properties   %(7.43b)
set(hf,'Color',[1 1 1],'Name','Short-distance word dependencies'); 
% short-distance word dependency computation	                   %(7.43c)
for m = 1:length(refer)          
    w1 = refer(m); w2 = probe(m); % gets reference and probe words
    % initializes the reference/probe word offset histogram 
    histprobe = zeros(1,9); 
    % loops over verses for computing co-occurrences  
    for k=1:length(limits)-1 
        % gets the next verse 
        theverse = wordsofverses(limits(k)+1:limits(k+1)-1); 
        % finds occurrences of the reference word, if there are any
        w1loc = find(theverse==w1); 
        % finds occurrences of the probe word, if there are any
        w2loc = find(theverse==w2);  
        % double-loop for computing all offsets between reference and probe
        for n=1:length(w1loc), for j=1:length(w2loc)
            index = w2loc(j)-w1loc(n); % computes the offset
            % updates histogram if offset is within the considered interval
            if (index>-5) && (index<5)
                histprobe(index+5) = histprobe(index+5)+1; end;
        end; end; % end of double-loop for offset computation 
    end; % end of loop for co-occurrence computation 
    subplot(2,2,m); hb = bar(-4:4,histprobe); % plots the histogram 
    temp = axis; axis([-5,5,0,temp(4)+1]); % sets axis limits 
    ylabel('Frequency'); % prints the corresponding labels 
    xlabel(sprintf('Location of "%s" with respect to "%s"',w2,w1));
end;
pause;

